//____________________________________________________________________________
/*!

\class   genie::flux::GSeaCORSIKAFileFlux.cxx

\brief   A flux driver to read incident muons and neutrinos from CORSIKA7 binary files

\author  Piotr Kalaczyński <piotr.kalaczynski@ncbj.gov.pl>
         NCBJ Warsaw

\created 25.12.2020

\cpright  Copyright (c) 2021, The KM3NeT Collaboration
          For the full text of the license see $GSEAGEN/LICENSE

*/
//____________________________________________________________________________

#include <fstream>
#include <cassert>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <unordered_map>

#include <TH2D.h>
#include <TH3D.h>
#include <TMath.h>
#include <TF1.h>
#include <TF3.h>
#include <TFile.h>

#ifdef _GENIEVERSIONGTEQ3__
#include "Framework/Conventions/Constants.h"
#include "Framework/Messenger/Messenger.h"
#include "Framework/ParticleData/PDGLibrary.h"
#else
#include "Conventions/Constants.h"
#include "Messenger/Messenger.h"
#include "PDG/PDGLibrary.h"
#endif

#include "SeaNuDrivers/GSeaCORSIKAFileFlux.h"


using namespace std;
using namespace genie;
using namespace genie::flux;
using namespace genie::constants;


//___________________________________________________________________________
GSeaCORSIKAFileFlux::GSeaCORSIKAFileFlux(GenParam * GenPar) :
GSeaFileFlux(GenPar)
{
  if(fGenPar->FluxFiles.size()>1){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  if(fGenPar->FluxFiles[0].FileNames.size()>1){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Only one input file is allowed, exiting...";
    exit(1);
  }

  for(unsigned i=0; i<fGenPar->NuList.size(); i++)fPdgCList->push_back(fGenPar->NuList[i]);

  LOG("GSeaCORSIKAFileFlux", pINFO) <<"-== Opening CORSIKA binary file: "<< fGenPar->FluxFiles[0].FileNames[0];
  CORSIKAFile.open(fGenPar->FluxFiles[0].FileNames[0]);
  if(!CORSIKAFile.is_open()){
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Error in opening file "<<fGenPar->FluxFiles[0].FileNames[0]<<", exiting...";
    exit(1);
  }

  CORSIKAFile.read((char*)&StdData,sizeof(StdData)); // (buffer,length)

  // check for the 32/64-bit architecture and standard/thinning format
  if ( 217433.0 < StdData[LenSubblStd+1] && StdData[LenSubblStd+1] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 32-bit file, standard format ==-";
     LBit = 32;
     LThi = 0;
  }
  else if ( 217433.0 < StdData[LenSubblStd+2] && StdData[LenSubblStd+2] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 64-bit file, standard format ==-";
     LBit = 64;
     LThi = 0;
  }
  else if ( 217433.0 < StdData[LenSubblThi+1] && StdData[LenSubblThi+1] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 32-bit file, thinning format ==-";
     LBit = 32;
     LThi = 1;
  }
  else if ( 217433.0 < StdData[LenSubblThi+2] && StdData[LenSubblThi+2] < 217433.2 ) {
     LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"-== 64-bit file, thinning format ==-";
     LBit = 64;
     LThi = 1;
  }
  LBit = (LBit-32) / 32;
  LenSubblock = LenSubblStd + LThi*39;

  CORSIKAFile.clear( );
  CORSIKAFile.seekg( 0, ios::beg ); // go back to the beginning of the file
  iRec=1,lp=1;
  PrimID=0;



  LOG("GSeaCORSIKAFileFlux", pNOTICE) << "Instantiating the CORSIKA7 file flux driver";

  LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"Reading the header ...";
  bool FullEvent=false;
  while(!FullEvent){ // iterate until a complete event is extracted
    if (lp >= LenSubblock*21) { // only read if all subblocks are read
      lp=1; // subblock index has to be reset
      // treating the data records according to their format (standard/thinning):
      if ( LThi == 1 ) { // long record (thinning)

         CORSIKAFile.read((char*)&ThiData,sizeof(ThiData));


         if ( LBit == 1 ) {
            CORSIKAFile.read((char*)&ZData,sizeof(ZData));
            ThiData[0] = 0.21345763838967e-41;
            for( int i=1; i<NumThi; i++ ) ThiData[i] = ThiData[i+LBit];
            ThiData[NumThi] = ZData[0];
         }

         for( int i=1; i<=NumStd; i++ ) StdData[i] = ThiData[i];
      }
      else if ( LThi == 0 ) { // short record (standard)

         CORSIKAFile.read((char*)&StdData,sizeof(StdData));

         if ( LBit == 1 ) CORSIKAFile.read((char*)&ZData,sizeof(ZData));

         StdData[0] = 0.21345763838967e-41;
         for( int i=NumStd; i<=NumThi; i++ ) ThiData[i] = 0.;
         for( int i=1; i<=NumStd; i++ ) ThiData[i] = StdData[i+LBit];
      }
      // to be sure we do not go too far
      if ( (iRec>fGenPar->MaxNRec) && (fGenPar->MaxNRec!=0) ) {
        LOG("GSeaCORSIKAFileFlux", pFATAL) << "Exceeded the maximum number of records, exiting...";
        exit(1);
      }
      iRec++;
    }
    else {

      LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading subblock " << ((lp-1)/LenSubblock+1) << " ==-";
      if ( (211285.2 < StdData[lp]) && (StdData[lp] < 211285.3) ) {       // RUNH = 211285.2812500000000000
         // double RunH=StdData[lp]               // normally RUNH = 211285
         nRun = StdData[lp+1];    // run number
         if (nRun == 0){
           LOG("GSeaCORSIKAFileFlux", pFATAL) << "CORSIKA file does not have a header, exiting...";
           exit(1);
         }
         fGenPar->CORSIKADate = StdData[lp+2];    // yymmdd of the run
         fGenPar->CORSIKAVer = StdData[lp+3];     // version of CORSIKA
         // double obslev = StdData[lp+5];          // observation level [cm]
         fGenPar->Alpha = StdData[lp+15];         // spectral index of E_primary
         fGenPar->EPrimMin = StdData[lp+16];      // minimal E_primary
         fGenPar->EPrimMax = StdData[lp+17];      // maximal E_primary
         fGenPar->EvMin = StdData[lp+20];         // minimal E_hadron -> ~ minimal neutrino energy
         // double EeMin = StdData[lp+21];          // minimal E_electron
         fGenPar->EmuMin = StdData[lp+22];        // minimal E_muon
         // double EgammaMin = StdData[lp+23];      // minimal E_gamma
         fGenPar->NTot = StdData[lp+92];          // number of generated showers
      }
      else if ( (217433.0 < StdData[lp]) && (StdData[lp] < 217433.1) ) {  // EVTH = 217433.0781250000000000
         PrimID = StdData[lp+2];                  // particle ID (CORSIKA scheme!)

         fGenPar->Primary=CorsikaToPdg(PrimID);
         fGenPar->PrimCtMin=TMath::Cos(kPi+StdData[lp+81]*kPi/180.);  // min cos(zen) is for the max zen
         fGenPar->PrimCtMax=TMath::Cos(kPi+StdData[lp+80]*kPi/180.); // max cos(zen) is for the min zen
         // int numSeeds = StdData[lp+12]; // Number of seeds
         fGenPar->CORSIKASeed[0]=StdData[lp+13]; // CORSIKA seed #1
         fGenPar->CORSIKASeed[1]=StdData[lp+16]; // CORSIKA seed #2
         switch((int)StdData[lp+74]) { // LE flag
           case 1 : fGenPar->CORSIKAHadronicModelLE = "GHEISHA"; break;
           case 2 : fGenPar->CORSIKAHadronicModelLE = "UrQMD"; break;
           case 3 : fGenPar->CORSIKAHadronicModelLE = "FLUKA"; break;
         }
         switch((int)StdData[lp+75]) { // HE flag
           case 0 : fGenPar->CORSIKAHadronicModelHE = "HDPM"; break;
           case 1 : fGenPar->CORSIKAHadronicModelHE = "VENUS"; break;
           case 2 : fGenPar->CORSIKAHadronicModelHE = "SIBYLL"; break;
           case 3 : fGenPar->CORSIKAHadronicModelHE = "QGSJET"; break;
           case 4 : fGenPar->CORSIKAHadronicModelHE = "DPMJET"; break;
           case 5 : fGenPar->CORSIKAHadronicModelHE = "NEXUS"; break;
           case 6 : fGenPar->CORSIKAHadronicModelHE = "EPOS"; break;
         }
         switch((int)StdData[lp+138]) { // SIBYLL interaction version flag
           case 0 : break; // SIBYLL not used
           case 1 : fGenPar->CORSIKAHadronicModelHE.append("-1.6"); break;
           case 2 : fGenPar->CORSIKAHadronicModelHE.append("-2.1"); break;
           case 3 : fGenPar->CORSIKAHadronicModelHE.append("-2.3c"); break;
           case 4 : fGenPar->CORSIKAHadronicModelHE.append("-2.3d"); break;
         }
         switch((int)StdData[lp+140]) { // QGSJET version flag
           case 0 : break; // QGSJET not used
           case 1 : fGenPar->CORSIKAHadronicModelHE.append("OLD"); break;
           case 2 : fGenPar->CORSIKAHadronicModelHE.append("01d"); break;
           case 3 : fGenPar->CORSIKAHadronicModelHE.append("-II"); break;
         }
         // other models have nothing to be added
      }
      else if ( (3397.3 < StdData[lp]) && (StdData[lp] < 3397.4) ) {      // EVTE =   3397.3918457031250000
         LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Event end (EVTE subblock) found (subblock #" << lp << " record #" << iRec << ") ==-";
         FullEvent = true; // end loop, full event extracted
         lp+=LenSubblock; // move to the next subblock
         break;
      }
      lp+=LenSubblock; // move to the next subblock
    }
  }

  CORSIKAFile.clear( );
  CORSIKAFile.seekg( 0, ios::beg ); // go back to the beginning of the file
  iRec=0,lp=1;
  PrimID=0;
  LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"Finished reading the header";
}

//___________________________________________________________________________
GSeaCORSIKAFileFlux::~GSeaCORSIKAFileFlux()
{

  CORSIKAFile.close();

}

//___________________________________________________________________________
vector<GSeaTrack> GSeaCORSIKAFileFlux::ReadFileEvent()
{

  vector<GSeaTrack> Tracks;

  int id = 0;

  double RT=-1.;
  if     (fGenPar->RTOpt.compare("proj")==0) RT = 0.;
  else if(fGenPar->RTOpt.compare("can")==0)  RT = sqrt(pow((fGenPar->Can2-fGenPar->Can1),2)+pow(fGenPar->Can3*2,2))/2.;
  assert(RT>=0.);

  double x=0.,y=0.,z=0.;
  double Xc=0.;
  double Yc=0.;
  double Zc=-fGenPar->SeaBottomRadius+fGenPar->Can1;  // in m
  double R=fGenPar->SeaBottomRadius+fGenPar->SiteDepth;

  RandomGen * rnd = RandomGen::Instance();

  GSeaTrack AuxTrack,AuxMother,AuxGrandMother,AuxMuonAddInfo;
  FullEvent = false;
  FullRecord = false;
  bool OnlyPrimaryTrack = true;
  while (OnlyPrimaryTrack){
    while (!FullEvent){ // iterate until a complete event is extracted
      if ((lp >= LenSubblock*21) || ((lp==1) && (iRec==0)))  { // read if all subblocks of the record are read
        
        if ( (iRec>fGenPar->MaxNRec) && (fGenPar->MaxNRec!=0) ) { // to be sure we do not go too far
          LOG("GSeaCORSIKAFileFlux", pFATAL) << "Exceeded the maximum number of records, exiting...";
          exit(1);
        }
        
        // if (iRec>0 && !FullEvent && !FullRecord) { // return Tracks if the event is bigger than one record
        //   FullRecord=true;
        //   if (Tracks.size()>1) {
        //     LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"Passing intermediate track list";
        //     return Tracks;
        //   }
        //   else if (Tracks.size()==0) LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"No tracks to pass";
        //   else if (Tracks.size()==1) LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"Empty record: no secondary reached the observation level in CORSIKA";
        // }
        // vector<GSeaTrack>().swap(Tracks); // clear memory from the previous record
        // FullRecord=false;
        
        LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading record " << iRec << " ==-";
        
        // treating the data records according to their format (standard/thinning):
        if ( LThi == 1 ) { // long record (thinning)

           CORSIKAFile.read((char*)&ThiData,sizeof(ThiData));


           if ( LBit == 1 ) {
              CORSIKAFile.read((char*)&ZData,sizeof(ZData));
              ThiData[0] = 0.21345763838967e-41;
              for( int i=1; i<NumThi; i++ ) ThiData[i] = ThiData[i+LBit];
              ThiData[NumThi] = ZData[0];
           }

           for( int i=1; i<=NumStd; i++ ) StdData[i] = ThiData[i];
        }
        else if ( LThi == 0 ) { // short record (standard)

           CORSIKAFile.read((char*)&StdData,sizeof(StdData));

           if ( LBit == 1 ) CORSIKAFile.read((char*)&ZData,sizeof(ZData));

           StdData[0] = 0.21345763838967e-41;
           for( int i=NumStd; i<=NumThi; i++ ) ThiData[i] = 0.;
           for( int i=1; i<=NumStd; i++ ) ThiData[i] = StdData[i+LBit];
        }
        
        lp=1; // subblock index has to be reset
        iRec++;
      }
      else {

        LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Reading subblock " << ((lp-1)/LenSubblock+1) << " ==-";
        if ( (211285.2 < StdData[lp]) && (StdData[lp] < 211285.3) ) {       // RUNH = 211285.2812500000000000
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Run header (RUNH subblock) found (subblock #" << lp << " record #" << iRec << ") ==-";

           // double RunH=StdData[lp]               // normally RUNH = 211285
           fGenPar->RunNu = StdData[lp+1];          // run number
           // float corsikaVer = StdData[lp+3];     // version of CORSIKA
           // double obslev = StdData[lp+5];        // observation level [cm]
           fGenPar->Alpha = StdData[lp+15];         // spectral index of E_primary
           fGenPar->EPrimMin = StdData[lp+16];      // minimal E_primary
           fGenPar->EPrimMax = StdData[lp+17];      // maximal E_primary
           fGenPar->EvMin = StdData[lp+20];         // minimal E_hadron -> ~ minimal neutrino energy
           // double EeMin = StdData[lp+21];        // minimal E_electron
           fGenPar->EmuMin = StdData[lp+22];        // minimal E_muon
           // double EgammaMin = StdData[lp+23];    // minimal E_gamma
           fGenPar->NTot = StdData[lp+92];          // number of generated showers
           if (fGenPar->RunNu == 0){
             LOG("GSeaCORSIKAFileFlux", pFATAL) << "CORSIKA file does not have a header, exiting...";
             exit(1);
           }

        }
        else if ( (217433.0 < StdData[lp]) && (StdData[lp] < 217433.1) ) {  // EVTH = 217433.0781250000000000
           LOG("GSeaCORSIKAFileFlux", pINFO) <<"-== Event header (EVTH subblock) found (shower #" << StdData[lp+1] << "/"<<fGenPar->NTot<<" ,record #" << iRec << ") ==-";
           vector<GSeaTrack>().swap(Tracks); // clear memory
           fDistaMax=0.; // resets the DistaMax to be evaluated for the next shower

           // Extracting selected information from the file:
           // for more details see CORSIKA Userguide: 10.2 Normal Particle Output
           // (p.124-139)
           // double EvtH=StdData[lp];                 // normally EVTH = 211285
           EventNumber = (unsigned int) StdData[lp+1];  // event (shower) number
           PrimID = StdData[lp+2];                      // particle ID (CORSIKA scheme!)
           EPrim = StdData[lp+3];                       // E_primary [GeV]
           // double StartAlt = StdData[lp+4];            // starting altitude [g/cm^2]
           // unsigned int NumT = StdData[lp+5];          // number of 1st target if fixed
           double XInt = 0.;                            // x coordinate of 1st interaction [cm]
           double YInt = 0.;                            // y coordinate of 1st interaction [cm]
                                                        // both x and y are by design 0
           double HInt = StdData[lp+6];                 // height of 1st interaction [cm] (z coord., negative if tracking starts at margin of atmosphere)
           double MomentumPrimX = StdData[lp+7];        // p_x [GeV/c] before interaction
           double MomentumPrimY = StdData[lp+8];        // p_y [GeV/c] before interaction
           double MomentumPrimZ = StdData[lp+9];        // p_z [GeV/c] before interaction (in -z direction!)
           // double Zenith = StdData[lp+10];             // zenith [rad]
           // double Azimuth = StdData[lp+11];            // azimuth [rad]
           MinEPrim = StdData[lp+58];                   // min E_primary [GeV]
           MaxEPrim = StdData[lp+59];                   // max E_primary [GeV]

           double DirXPrim=MomentumPrimX/(TMath::Sqrt(TMath::Power(MomentumPrimX,2)+TMath::Power(MomentumPrimY,2)+TMath::Power(MomentumPrimZ,2)));
           double DirYPrim=MomentumPrimY/(TMath::Sqrt(TMath::Power(MomentumPrimX,2)+TMath::Power(MomentumPrimY,2)+TMath::Power(MomentumPrimZ,2)));
           double DirZPrim=MomentumPrimZ/(TMath::Sqrt(TMath::Power(MomentumPrimX,2)+TMath::Power(MomentumPrimY,2)+TMath::Power(MomentumPrimZ,2)));

           // rotate from CORSIKA to KM3NeT coordinates:
           AuxTrack.V1=XInt/100.;  // x position [m]
           AuxTrack.V2=YInt/100.;  // y position [m]
           AuxTrack.V3=-HInt/100.; // z position [m]
           AuxTrack.D1=TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * DirXPrim - TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * DirYPrim; // x direction
           AuxTrack.D2=TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * DirXPrim + TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * DirYPrim; // y direction
           AuxTrack.D3=-1. * DirZPrim; // z direction

           AuxTrack.E=EPrim;
           AuxTrack.T=0; // time counter (starts at 1st interaction)
           AuxTrack.PdgCode=CorsikaToPdg(PrimID);

           AuxTrack.Id = id++;
           AuxTrack.MotherId = -1;
           AuxTrack.Status = 200;      // i.e. primary particle
           AuxTrack.V1 = TMath::Abs(AuxTrack.V3/AuxTrack.D3)*-1*AuxTrack.D1;
           AuxTrack.V2 = TMath::Abs(AuxTrack.V3/AuxTrack.D3)*-1*AuxTrack.D2;
           AuxTrack.Length = 0.;
           AuxTrack.GenCounter = 0;
           Tracks.push_back(AuxTrack);

        }
        else if ( (3397.3 < StdData[lp]) && (StdData[lp] < 3397.4) ) {      // EVTE =   3397.3918457031250000
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Event end (EVTE subblock) found ==-";

           // all the following has to be computed at the end of the event, because
           // fDistaMax is evaluated based on positions of all the muons in the shower
           fDistaMax += 100.; // just a safety margin

           // generate the primary vertex
           double b  = Tracks[0].D1*(fX0-Xc) + Tracks[0].D2*(fY0-Yc) + Tracks[0].D3*(fZ0-Zc);
           double c  = TMath::Power(fX0-Xc,2)+TMath::Power(fY0-Yc,2)+TMath::Power(fZ0-Zc,2)-R*R;
           double RL = TMath::Abs(-b-TMath::Sqrt(b*b-c));
           x = 0.;
           y = 0.;
           z = 0.;
           if (RT==0.) {
             double height = (fGenPar->Can2-fGenPar->Can1+fGenPar->HBedRock) + fDistaMax;
             double radius = fGenPar->Can3 + fDistaMax;
             double AreaTop  = kPi * TMath::Power(radius,2) * TMath::Abs(Tracks[0].D3) ;
             double AreaSide = 2. * height * radius * TMath::Sqrt( 1. - TMath::Power(Tracks[0].D3,2) );
             fAgen = AreaTop + AreaSide;
             if ( rnd->RndFlux().Rndm() < AreaTop / fAgen ) {
               double r   = radius * TMath::Sqrt( rnd->RndFlux().Rndm() );
               double phi = 2. * kPi * rnd->RndFlux().Rndm();
               x = r*TMath::Sin(phi);
               y = r*TMath::Cos(phi);
               z = Tracks[0].D3 > 0 ? -height/2. : height/2.;
             }
             else {
               double zaux = height * (rnd->RndFlux().Rndm()-0.5);
               double yaux = 2. * radius * (rnd->RndFlux().Rndm()-0.5);
               double xaux = -TMath::Sqrt( radius*radius - yaux*yaux );
               TVector3 v(xaux,yaux,zaux);
               double phidir = TMath::ATan2( Tracks[0].D2, Tracks[0].D1 );
               v.RotateZ( phidir );
               x = v.X();
               y = v.Y();
               z = v.Z();
             }
             x += -RL * Tracks[0].D1;
             y += -RL * Tracks[0].D2;
             z += -RL * Tracks[0].D3;
           }
           else {
             RT+=fDistaMax;
             fAgen = kPi*TMath::Power(RT,2);
             x += -RL * Tracks[0].D1;                  // we have to multiply RL times the arrival direction cosines
             y += -RL * Tracks[0].D2;
             z += -RL * Tracks[0].D3;

             TVector3 vec(x,y,z);                      // vector towards selected point
             TVector3 dvec1 = vec.Orthogonal();        // orthogonal vector
             TVector3 dvec2 = dvec1;                   // second orthogonal vector
             dvec2.Rotate(-kPi/2.,vec);                // rotate second vector by 90deg, now forming a new orthogonal cartesian coordinate system
             double psi = 2.*kPi* rnd->RndFlux().Rndm(); // rndm angle [0,2pi]
             double random = rnd->RndFlux().Rndm();    // rndm number  [0,1]
             dvec1.SetMag(TMath::Sqrt(random)*RT*TMath::Cos(psi));
             dvec2.SetMag(TMath::Sqrt(random)*RT*TMath::Sin(psi));
             x += dvec1.X() + dvec2.X();
             y += dvec1.Y() + dvec2.Y();
             z += dvec1.Z() + dvec2.Z();
           }
           x += fX0;
           y += fY0;
           z += fZ0;

           for (auto& Track : Tracks) {
             Track.V1 += x;
             Track.V2 += y;
             Track.V3 += z;
             // for muons and neutrinos, we translated the starting point from the generation area up to the sea level
             if ( (Track.Status==-1) && (pdg::IsNeutrino(TMath::Abs(Track.PdgCode)) || pdg::IsMuon(TMath::Abs(Track.PdgCode))) ) {
               double bmu  = Track.D1*(Track.V1-Xc)+ Track.D2*(Track.V2-Yc) +Track.D3*(Track.V3-Zc);
               double cmu  = TMath::Power(Track.V1-Xc,2)+TMath::Power(Track.V2-Yc,2)+TMath::Power(Track.V3-Zc,2)-R*R;
               double RLmu = -bmu-TMath::Sqrt(bmu*bmu-cmu);
               if (fGenPar->CustomRotation[0]!=0. || fGenPar->CustomRotation[1]!=0. || fGenPar->CustomRotation[2]!=0.) {
                 double dir_x = Track.D1;
                 double dir_y = Track.D2;
                 double dir_z = Track.D3;
                 // rotation according to https://www.sciencedirect.com/topics/engineering/direction-cosine-matrix (eq. 11.119)
                 Track.D1 = (
                   dir_x*TMath::Cos(fGenPar->CustomRotation[0])*TMath::Cos(fGenPar->CustomRotation[1])+
                   dir_y*TMath::Sin(fGenPar->CustomRotation[0])*TMath::Cos(fGenPar->CustomRotation[1])-
                   dir_z*TMath::Sin(fGenPar->CustomRotation[1])
                 );
                 Track.D2 = (
                   dir_x*(TMath::Cos(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[1])*TMath::Sin(fGenPar->CustomRotation[2])-TMath::Sin(fGenPar->CustomRotation[0])*TMath::Cos(fGenPar->CustomRotation[2]))+
                   dir_y*(TMath::Sin(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[1])*TMath::Sin(fGenPar->CustomRotation[2])+TMath::Cos(fGenPar->CustomRotation[0])*TMath::Cos(fGenPar->CustomRotation[2]))+
                   dir_z*TMath::Cos(fGenPar->CustomRotation[1])*TMath::Sin(fGenPar->CustomRotation[2])
                 );
                 Track.D3 = (
                   dir_x*(TMath::Cos(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[1])*TMath::Cos(fGenPar->CustomRotation[2])+TMath::Sin(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[2]))+
                   dir_y*(TMath::Sin(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[1])*TMath::Cos(fGenPar->CustomRotation[2])-TMath::Cos(fGenPar->CustomRotation[0])*TMath::Sin(fGenPar->CustomRotation[2]))+
                   dir_z*TMath::Cos(fGenPar->CustomRotation[1])*TMath::Cos(fGenPar->CustomRotation[2])
                 );
               }
               Track.V1 += RLmu * Track.D1;
               Track.V2 += RLmu * Track.D2;
               Track.V3 += RLmu * Track.D3;

             }
           }

           // Compute the weights
           // (accodring to https://simulation.pages.km3net.de/taglist/taglist.pdf)

           double waux = fAgen; // w1

           double ITheta = 2. * kPi * TMath::Abs(fGenPar->PrimCtMax - fGenPar->PrimCtMin);
           double Gamma = TMath::Abs(fGenPar->Alpha);
           double IEprim;
           if (Gamma == 1) IEprim = TMath::Log(MaxEPrim / MinEPrim);
           else IEprim = (TMath::Power(MaxEPrim, 1. - Gamma) - TMath::Power(MinEPrim, 1. - Gamma)) / (1. - Gamma);

           fGenWeight = waux * ITheta * IEprim * TMath::Power(EPrim, Gamma) * fGenPar->TGen; // w2

           fEvtWeight = fGenWeight * CRFlux(EPrim, PrimID, fGenPar->CRModel); // w3

           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"waux: "<<waux;
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fGenWeight: "<<fGenWeight;
           LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fEvtWeight: "<<fEvtWeight;

           FullEvent = true;  // end loop, full event extracted
           lp+=LenSubblock;   // move to the next subblock
           break;
        }
        else if ( (3301.3 < StdData[lp]) && (StdData[lp] < 3301.4) ) {      // RUNE =   3301.3325195312500000
          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Run end (RUNE subblock) found (record #" << iRec << ") ==-";
          FullEvent = true;   // end loop, full event extracted
          FullRun = true;     // full run extracted
          OnlyPrimaryTrack = false;
          lp+=LenSubblock;    // move to the next subblock
          vector<GSeaTrack>().swap(Tracks); // clear memory
          break;
        }
        else if ( (52815.2 < StdData[lp]) && (StdData[lp] < 52815.3) ) {    // LONG =  52815.2968750000000000
          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Analyzing LONG subblock ... ==-";
        }
        else {                                                              // Particle subblock

          LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== Analyzing particle subblock (record #" << iRec <<") ... ==-";

          for (int j = 0; j < (4*39-1) ; j++) { // iterating over particles (up to 4x39)

            // Extracting selected information from the file:
            // for more details see CORSIKA Userguide: 10.2 Normal Particle Output
            // (p.124-139)
            int ParticleID = int(StdData[lp+7*j])/1000; // pID (CORSIKA scheme!)
            int GenCounter = int(StdData[lp+7*j])%1000; // generation counter
            if (ParticleID==0){
              LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"-== End of particle subblock (len="<<(j+1)<<") ==-";
              //cout << endl;
              break; // no more particles in the subblock
            }

            double MomentumX = StdData[lp+1+7*j];       // p_x [GeV/c]
            double MomentumY = StdData[lp+2+7*j];       // p_y [GeV/c]
            double MomentumZ = StdData[lp+3+7*j];       // p_z [GeV/c]
            double PosXUnrotated = StdData[lp+4+7*j];   // x coordinate [cm]
                                                        // or
                                                        // extended electromagnetic generation counter of final particle
            double PosYUnrotated = StdData[lp+5+7*j];   // y coordinate [cm]
                                                        // or
                                                        // penetration depth in [g/cm2]
            double PosZOrTime = StdData[lp+6+7*j];      // t [ns] since first interaction
                                                        // or
                                                        // z [cm] of muon production point (for additional muon info (MUADDI))

            // calculate the directions still in CORSIKA coordinate system:
            double DirX=MomentumX/(TMath::Sqrt(TMath::Power(MomentumX,2)+TMath::Power(MomentumY,2)+TMath::Power(MomentumZ,2)));
            double DirY=MomentumY/(TMath::Sqrt(TMath::Power(MomentumX,2)+TMath::Power(MomentumY,2)+TMath::Power(MomentumZ,2)));
            double DirZ=MomentumZ/(TMath::Sqrt(TMath::Power(MomentumX,2)+TMath::Power(MomentumY,2)+TMath::Power(MomentumZ,2)));

            // rotate pos and dir from CORSIKA to KM3NeT coordinates:
            double PosX = (TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * PosXUnrotated - TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * PosYUnrotated)/100.;  // x position [m]
            double PosY = (TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * PosXUnrotated + TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * PosYUnrotated)/100.;  // y position [m]
            // PosZ is handled separately in cases below
            AuxTrack.D1=TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * DirX - TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * DirY; // x direction
            AuxTrack.D2=TMath::Sin(0.5*kPi + fGenPar->SiteDeclination) * DirX + TMath::Cos(0.5*kPi + fGenPar->SiteDeclination) * DirY; // y direction
            AuxTrack.D3=-1. * DirZ; // z direction

            AuxTrack.E=TMath::Sqrt(TMath::Power(MomentumX,2)+TMath::Power(MomentumY,2)+TMath::Power(MomentumZ,2)+TMath::Power(CorsikaRestMass(ParticleID),2));

            if (ParticleID>=0) AuxTrack.PdgCode=CorsikaToPdg(ParticleID);
            else AuxTrack.PdgCode=CorsikaToPdg(-1*ParticleID); // (grand-)mother

            if ( ParticleID==5 || ParticleID==6 ) AuxTrack.Status = -1;                 // muon
            // uncomment for neutrinos to be included for propagation:
            // else if (ParticleID==66 || ParticleID==67 || ParticleID==68 || ParticleID==68 || ParticleID==133  || ParticleID==134) AuxTrack.Status = -1;     // neutrino
            else if ( ParticleID==85 || ParticleID==86 ) AuxTrack.Status = -22;           // decaying mu at start (muon that does NOT reach the observation level!)
            else if ( ParticleID==95 || ParticleID==96 ) AuxTrack.Status = -23;         // decaying mu at end   (muon that does NOT reach the observation level!)
            else if ( ParticleID==71 ) AuxTrack.Status = -24;                           // eta -> 2 gamma
            else if ( ParticleID==72 ) AuxTrack.Status = -25;                           // eta -> 3 pi0
            else if ( ParticleID==73 ) AuxTrack.Status = -26;                           // eta -> pi+ pi- pi0
            else if ( ParticleID==74 ) AuxTrack.Status = -27;                           // eta -> pi+ pi- gamma
            else if ( ParticleID==9900 ) AuxTrack.Status = -28;                         // Cherenkov photons on particle output file
            else AuxTrack.Status = -21;                                                 // all other parents/muaddi
            
            AuxTrack.Length = 0.;
            AuxTrack.T = 0.;
            AuxTrack.GenCounter = 0;
            
            // the parent relation: primary -> ... -> grandmother -> mother -> mu add. info (mu at production point) -> muon at sea
            if (ParticleID==75 || ParticleID==76 || ParticleID==85 || ParticleID==86){ // Additional muon information
              AuxMuonAddInfo=AuxTrack;
              AuxMuonAddInfo.V1 = PosX;               // x coordinate [m] (creation point)
              AuxMuonAddInfo.V2 = PosY;               // y coordinate [m] (creation point)
              AuxMuonAddInfo.V3 = PosZOrTime/100.;    // z coordinate [m] (creation point)
              AuxMuonAddInfo.GenCounter = GenCounter; // due to the lack of more fitting fields, hadronic generation counter is saved in usr
                                                      // If mu is produced after photonuclear interaction or was created by
                                                      // mu pair production +500 is added in CORSIKA
            }
            else if (ParticleID<0){ // parent particles
              if (PosZOrTime<0){ // grandmother
                AuxGrandMother=AuxTrack;
                AuxGrandMother.V1 = 0.;                   // x coordinate is not available
                AuxGrandMother.V2 = 0.;                   // y coordinate is not available
                AuxGrandMother.V3 = PosZOrTime*-1./100.;  // z coordinate [m] is flipped
                AuxGrandMother.GenCounter = -GenCounter;  // electromagnetic generation counter of MOTHER!
                AuxTrack.GenCounter = PosXUnrotated;      // extended electromagnetic generation counter of secondary
                AuxTrack.PenDepth = PosYUnrotated;        // penetration depth [g/cm^2]
              }
              else{ // mother
                AuxMother=AuxTrack;
                AuxMother.V1 = PosX;                // x coordinate [m]
                AuxMother.V2 = PosY;                // y coordinate [m]
                AuxMother.V3 = PosZOrTime/100.;     // z coordinate [m]
                AuxMother.GenCounter = -GenCounter; // hadronic generation counter of mother
              }
            }
            else { // secondary particle or decaying muon at end (CORSIKA id 95/96)

              AuxTrack.V1 = PosX;             // x coordinate
              AuxTrack.V2 = PosY;             // y coordinate
              AuxTrack.V3 = 0.;               // z=obslev
              AuxTrack.T = PosZOrTime;        // t [ns] since first interaction

              // if it's a muon, we can calculate the track lenght:
              if ( pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode)) ) AuxTrack.Length = TMath::Sqrt(TMath::Power((AuxTrack.V1-AuxMuonAddInfo.V1),2)+TMath::Power((AuxTrack.V2-AuxMuonAddInfo.V2),2)+TMath::Power((AuxTrack.V3-AuxMuonAddInfo.V3),2));

              if ( // only the requested tracks are picked for later propagation
                  (fGenPar->SaveNu && pdg::IsNeutrino(TMath::Abs(AuxTrack.PdgCode))) || // neutrino tracks
                  (fGenPar->SaveMu && pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode))) ||     // muon tracks
                  (fGenPar->SaveOther)                                                  // tracks from other paricles
              ) {

                // setting the IDs in chronological order
                AuxGrandMother.Id = id++;
                AuxMother.Id = id++;
                if ( pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode)) ) AuxMuonAddInfo.Id = id++; // only increment if there is a muon
                AuxTrack.Id = id++;

                // setting the mother IDs
                AuxGrandMother.MotherId = Tracks[0].Id;   // corresponding primary
                AuxMother.MotherId = AuxGrandMother.Id;
                if ( pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode)) ) {
                  AuxMuonAddInfo.MotherId = AuxMother.Id; // mother particle is parent of the muon at production point
                  AuxTrack.MotherId = AuxMuonAddInfo.Id;
                }
                else AuxTrack.MotherId = AuxMother.Id;

                // passing the tracks
                Tracks.push_back(AuxGrandMother);
                Tracks.push_back(AuxMother);
                if ( pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode)) ) Tracks.push_back(AuxMuonAddInfo);
                Tracks.push_back(AuxTrack);
              }

              // only muons that have reached the observation level (sea) and have sufficient range
              // to reach the can are allowed to affect the drawing area:
              if ( ( pdg::IsMuon(TMath::Abs(AuxTrack.PdgCode)) ) && ( AuxTrack.Status == -1 ) ) {
                double fDistaCan = fabs((fGenPar->SiteDepth-fGenPar->Can2-AuxTrack.V3)/AuxTrack.D3); // fGenPar->Can2 = Zmax
                double Range = pow(10.,fGenPar->RangeSW.Eval(log10(AuxTrack.E)))/(fGenPar->RhoSW);
                if (Range>=fDistaCan) fDistaMax = TMath::Max(
                  fDistaMax,
                  TMath::Sqrt(
                    (
                      TMath::Power(AuxTrack.V2*Tracks[0].V3-AuxTrack.V3*Tracks[0].V2,2)+
                      TMath::Power(AuxTrack.V3*Tracks[0].V1-AuxTrack.V1*Tracks[0].V3,2)+
                      TMath::Power(AuxTrack.V1*Tracks[0].V2-AuxTrack.V2*Tracks[0].V1,2)
                    )
                   /
                    (
                      TMath::Power(Tracks[0].V1,2)+
                      TMath::Power(Tracks[0].V2,2)+
                      TMath::Power(Tracks[0].V3,2)
                    )
                  )
                );

                LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fDistaMax:"<<fDistaMax;
                LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"fDistaMax limit:"<<TMath::Sqrt(TMath::Power(fGenPar->Can2-fGenPar->Can1,2)+TMath::Power(2.*fGenPar->Can3,2));
              }

            }
          }
        }
        lp+=LenSubblock; // move to the next subblock
      }
    }
    if (Tracks.size()>1) {
      LOG("GSeaCORSIKAFileFlux", pDEBUG) <<"Passing track list";
      OnlyPrimaryTrack=false;
      LOG("GSeaCORSIKAFileFlux", pDEBUG) << "Tracks.size(): "<<Tracks.size();
      return Tracks;
    }
    else if (Tracks.size()==0){
      LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"No tracks to pass";
      OnlyPrimaryTrack=false;
      
      return Tracks;
    }
    else if (Tracks.size()==1){
      LOG("GSeaCORSIKAFileFlux", pNOTICE) <<"Empty event: no secondary reached the observation level in CORSIKA";
      FullEvent = false;
      OnlyPrimaryTrack=true;
    }
  }

}

//___________________________________________________________________________
int GSeaCORSIKAFileFlux::CorsikaToPdg(int CORSIKAId){
  std::unordered_map<int, int> IDS = {
        // 0 is NOT a valid PDG ID, but it is used for options not covered by
        // the PDG scheme

        //{CORSIKA, PDG}        // PARTICLE NAME
          { 1, 22 },            // photon
          { 2, -11 },           // positron
          { 3, 11 },            // electron

          { 5, -13 },           // muon +
          { 6, 13 },            // muon -
          { 7, 111 },           // pion0
          { 8, 211 },           // pion+
          { 9, -211 },          // pion-
          { 10, 130 },          // kaon0_L
          { 11, 321 },          // kaon+
          { 12, -321 },         // kaon-
          { 13, 2112 },         // neutron
          { 14, 2212 },         // proton
          { 15, -2212 },        // antiproton
          { 16, 310 },          // kaon0_S
          { 17, 221 },          // eta
          { 18, 3122 },         // Lambda
          { 19, 3222 },         // Sigma+
          { 20, 3212 },         // Sigma0
          { 21, 3112 },         // Sigma-
          { 22, 3322 },         // Xi0
          { 23, 3312 },         // Xi-
          { 24, 3334 },         // Omega-
          { 25, -2112 },        // antineutron
          { 26, -3122 },        // antiLambda
          { 27, -3222 },        // antiSigma-
          { 28, -3212 },        // antiSigma0
          { 29, -3112 },        // antiSigma+
          { 30, -3322 },        // antiXi0
          { 31, -3312 },        // antiXi+
          { 32, -3334 },        // antiOmega+

          { 48, 331 },          // eta'
          { 49, 333 },          // phi
          { 50, 223 },          // omega
          { 51, 113 },          // rho0 (assuming 770 state)
          { 52, 213 },          // rho+ (assuming 770 state)
          { 53, -213 },         // rho- (assuming 770 state)
          { 54, 2224 },         // Delta++
          { 55, 2214 },         // Delta+
          { 56, 2114 },         // Delta0
          { 57, 1114 },         // Delta-
          { 58, -2224 },        // antiDelta--
          { 59, -2214 },        // antiDelta-
          { 60, -2114 },        // antiDelta0
          { 61, -1114 },        // antiDelta+
          { 62, 313 },          // kaon*0 (assuming 892 state)
          { 63, 323 },          // kaon*+ (assuming 892 state)
          { 64, -323 },         // antikaon*-
          { 65, -313 },         // antikaon*0
          { 66, 12 },           // electron neutrino
          { 67, -12 },          // antielectron neutrino
          { 68, 14 },           // muon neutrino
          { 69, -14 },          // antimuon neutrino

          { 71, 221 },          // eta -> 2 gamma         | status=-24
          { 72, 221 },          // eta -> 3 pi0           | status=-25
          { 73, 221 },          // eta -> pi+ pi- pi0     | status=-26
          { 74, 221 },          // eta -> pi+ pi- gamma   | status=-27
          { 75, -13 },          // muon+ add. info        | status=-21
          { 76, 13 },           // muon- add. info        | status=-21

          { 85, -13 },          // decaying mu+ at start  | status=-22
          { 86, 13 },           // decaying mu- at start  | status=-22

          { 95, -13 },          // decaying mu+ at end    | status=-23
          { 96, 13 },           // decaying mu- at end    | status=-23

          { 116, 421 },         // D0
          { 117, 411 },         // D+
          { 118, -411 },        // antiD-
          { 119, -421 },        // antiD0
          { 120, 431 },         // D_s+
          { 121, -431 },        // antiD_s-
          { 122, 441 },         // eta_c (assuming always eta_c (1S))
          { 123, 423 },         // D*0 (assuming 2007 state)
          { 124, 413 },         // D*+ (assuming 2010 state)
          { 125, -413 },        // antiD*-
          { 126, -423 },        // antiD*0
          { 127, 433 },         // D_s*-
          { 128, -433 },        // antiD_s*-

          { 130, 443 },         // J/psi
          { 131, -15 },         // tau+
          { 132, 15 },          // tau-
          { 133, 16 },          // nutau
          { 134, -16 },         // antinutau

          { 137, 4122 },        // Lambda_c+
          { 138, 4232 },        // Xi_c+
          { 139, 4132 },        // Xi_c0
          { 140, 4222 },        // Sigma_c++
          { 141, 4212 },        // Sigma_c+
          { 142, 4112 },        // Sigma_c0
          { 143, 4322 },        // Xi_c'+
          { 144, 4312 },        // Xi_c'0
          { 145, 4332 },        // Omega_c0

          { 149, -4122 },       // antiLambda_c-
          { 150, -4232 },       // antiXi_c-
          { 151, -4132 },       // antiXi_c0
          { 152, -4222 },       // antiSigma_c--
          { 153, -4212 },       // antiSigma_c-
          { 154, -4112 },       // antiSigma_c0
          { 155, -4322 },       // antiXi_c'-
          { 156, -4312 },       // antiXi_c'0
          { 157, -4332 },       // antiOmega_c0

          { 161, 4224 },        // Sigma_c*++
          { 162, 4214 },        // Sigma_c*+
          { 163, 4114 },        // Sigma_c*0

          { 171, -4224 },       // antiSigma_c*--
          { 172, -4214 },       // antiSigma_c*-
          { 173, -4114 },       // antiSigma_c*0

          { 176, 511 },         // B0
          { 177, 521 },         // B+
          { 178, -521 },        // antiB-
          { 179, -511 },        // antiB0
          { 180, 531 },         // B_s0
          { 181, -531 },        // antiB_s0
          { 182, 541 },         // B_c+
          { 183, -541 },        // antiB_c-
          { 184, 5122 },        // Lambda_b0
          { 185, 5112 },        // Sigma_b-
          { 186, 5222 },        // Sigma_b+
          { 187, 5232 },        // Xi_b0
          { 188, 5132 },        // Xi_b-
          { 189, 5332 },        // Omega_b-
          { 190, -5122 },       // antiLambda_b0
          { 191, -5112 },       // antiSigma_b+
          { 192, -5222 },       // antiSigma_b-
          { 193, -5232 },       // antiXi_b0
          { 194, -5132 },       // antiXi_b+
          { 195, -5332 },       // antiOmega_b+

          // Nuclei examples:    A x 100 + Z
          { 101, 1000010010 },  // 'hydrogen'
          { 201, 1000010020 },  // 'deuteron'},
          { 301, 1000010030 },  // 'tritium',
          { 302, 1000020030 },  // 'helium3',
          { 402, 1000020040 },  // 'alpha',
          { 703, 1000030070 },  // 'lithium',
          { 904, 1000040090 },  // 'beryllium',
          { 1105, 1000050110 }, // 'boron',
          { 1206, 1000060120 }, // 'carbon',
          { 1407, 1000070140 }, // 'nitrogen',
          { 1608, 1000080160 }, // 'oxygen',
          { 2713, 1000130270 }, // 'aluminium',
          { 2814, 1000140280 }, // 'silicon',
          { 3216, 1000160320 }, // 'sulfur',
          { 4020, 1000200400 }, // 'calcium',
          { 5626, 1000260560 }, // 'iron',
          { 5828, 1000280580 }, // 'nickel',

          { 9900, 1 },    // Cherenkov photons on particle output file
                          // (NOT there without the CERENKOV option, which )
                          // (is not used for KM3NeT CORSIKA simulations)
                          // | status=-28
  };

  if ((CORSIKAId < 200) || (CORSIKAId == 9900)) return IDS[CORSIKAId]; // covered above
  else {                      // nuclei
    int nZ = CORSIKAId%100;
    int nA = (CORSIKAId-nZ)/100;
    char out_id[10];
    sprintf(out_id, "100%03d%03d0", nZ, nA);
    int out_id_int = std::stoi (out_id,nullptr,10);
    return out_id_int;
  }
}

//___________________________________________________________________________
double GSeaCORSIKAFileFlux::CorsikaRestMass(int CORSIKAId){
  std::unordered_map<int, double> Masses = {
        // All masses are taken from PDG booklet 2020 (https://pdg.lbl.gov)
        // unless noted otherwise
        // neutrino masses are neglected

        //{CORSIKA, rest mass [GeV/c^2]}    // PARTICLE NAME
          { 1, 0. },                        // photon
          { 2, 0.0005109989461 },           // positron
          { 3, 0.0005109989461 },           // electron

          { 5, 0.1056583745 },              // muon +
          { 6, 0.1056583745 },              // muon -
          { 7, 0.1349768 },                 // pion0
          { 8, 0.13957039 },                // pion+
          { 9, 0.13957039 },                // pion-
          { 10, 0.497611 },                 // kaon0_L
          { 11, 0.493677 },                 // kaon+
          { 12, 0.493677 },                 // kaon-
          { 13, 0.939565413 },              // neutron
          { 14, 0.938272081 },              // proton
          { 15, 0.938272081 },              // antiproton
          { 16, 0.497611 },                 // kaon0_S
          { 17, 0.547862 },                 // eta
          { 18, 1.115683 },                 // Lambda
          { 19, 1.18937 },                  // Sigma+
          { 20, 1.192642 },                 // Sigma0
          { 21, 1.197449 },                 // Sigma-
          { 22, 1.31486 },                  // Xi0
          { 23, 1.32171 },                  // Xi-
          { 24, 1.67245 },                  // Omega-
          { 25, 0.939565413 },              // antineutron
          { 26, 1.115683 },                 // antiLambda
          { 27, 1.18937 },                  // antiSigma-
          { 28, 1.192642 },                 // antiSigma0
          { 29, 1.197449 },                 // antiSigma+
          { 30, 1.31486 },                  // antiXi0
          { 31, 1.32171 },                  // antiXi+
          { 32, 1.67245 },                  // antiOmega+

          { 48, 0.95778 },                  // eta' (assuming 958 state)
          { 49, 1.019461 },                 // phi (assuming 1020 state)
          { 50, 0.78265 },                  // omega (assuming 782 state)
          { 51, 0.77526 },                  // rho0 (assuming 770 state)
          { 52, 0.77526 },                  // rho+ (assuming 770 state)
          { 53, 0.77526 },                  // rho- (assuming 770 state)
          { 54, 2224 },                     // Delta++
          { 55, 1.210 },                    // Delta+ (1232 state)
          { 56, 2114 },                     // Delta0
          { 57, 1.610 },                    // Delta- (1620 state)
          { 58, -2224 },                    // antiDelta--
          { 59, 1.210 },                    // antiDelta-
          { 60, -2114 },                    // antiDelta0
          { 61, 1.610 },                    // antiDelta+
          { 62, 0.89555 },                  // kaon*0 (assuming 892 state and hadroproduction)
          { 63, 0.89166 },                  // kaon*+ (assuming 892 state and hadroproduction)
          { 64, 0.89166 },                  // antikaon*-
          { 65, 0.89555 },                  // antikaon*0
          { 66, 0. },                       // electron neutrino
          { 67, 0. },                       // antielectron neutrino
          { 68, 0. },                       // muon neutrino
          { 69, 0. },                       // antimuon neutrino

          { 71, 0.547862 },                 // eta -> 2 gamma
          { 72, 0.547862 },                 // eta -> 3 pi0
          { 73, 0.547862 },                 // eta -> pi+ pi- pi0
          { 74, 0.547862 },                 // eta -> pi+ pi- gamma
          { 75, 0.1056583745 },             // muon+ add. info
          { 76, 0.1056583745 },             // muon- add. info

          { 85, 0.1056583745 },             // decaying mu+ at start
          { 86, 0.1056583745 },             // decaying mu- at start

          { 95, 0.1056583745 },             // decaying mu+ at end
          { 96, 0.1056583745 },             // decaying mu- at end

          { 116, 1.86483 },                 // D0
          { 117, 1.86965 },                 // D+
          { 118, 1.86965 },                 // antiD-
          { 119, 1.86483 },                 // antiD0
          { 120, 1.96834 },                 // D_s+
          { 121, 1.96834 },                 // antiD_s-
          { 122, 2.9839 },                  // eta_c (assuming always eta_c (1S))
          { 123, 2.00685 },                 // D*0 (assuming 2007 state)
          { 124, 2.01026 },                 // D*+ (assuming 2010 state)
          { 125, 2.01026 },                 // antiD*-
          { 126, 2.00685 },                 // antiD*0
          { 127, 2.1122 },                  // D_s*-
          { 128, 2.1122 },                  // antiD_s*+

          { 130, 3.096900 },                // J/psi
          { 131, 1.77686 },                 // tau+
          { 132, 1.77686 },                 // tau-
          { 133, 0. },                      // nutau
          { 134, 0. },                      // antinutau

          { 137, 2.28646 },                 // Lambda_c+
          { 138, 2.46794 },                 // Xi_c+
          { 139, 2.47090 },                 // Xi_c0
          { 140, 2.45397 },                 // Sigma_c++
          { 141, 2.4529 },                  // Sigma_c+ (2455 state)
          { 142, 2.45375 },                 // Sigma_c0
          { 143, 2.5784 },                  // Xi_c'+
          { 144, 2.5792 },                  // Xi_c'0
          { 145, 2.6952 },                  // Omega_c0

          { 149, 2.28646 },                 // antiLambda_c-
          { 150, 2.46794 },                 // antiXi_c-
          { 151, 2.47090 },                 // antiXi_c0
          { 152, 2.45397 },                 // antiSigma_c--
          { 153, 2.4529 },                  // antiSigma_c-
          { 154, 2.45375 },                 // antiSigma_c0
          { 155, 2.5784 },                  // antiXi_c'-
          { 156, 2.5792 },                  // antiXi_c'0
          { 157, 2.6952 },                  // antiOmega_c0

          { 161, 2.5179 },                  // Sigma_c*++ from https://en.wikipedia.org/wiki/List_of_baryons
          { 162, 2.5175 },                  // Sigma_c*+  from https://en.wikipedia.org/wiki/List_of_baryons
          { 163, 2.5188 },                  // Sigma_c*0  from https://en.wikipedia.org/wiki/List_of_baryons

          { 171, 2.5179 },                  // antiSigma_c*--  from https://en.wikipedia.org/wiki/List_of_baryons
          { 172, 2.5175 },                  // antiSigma_c*-  from https://en.wikipedia.org/wiki/List_of_baryons
          { 173, 2.5188 },                  // antiSigma_c*0  from https://en.wikipedia.org/wiki/List_of_baryons

          { 176, 5.27965 },                 // B0
          { 177, 5.27934 },                 // B+
          { 178, 5.27934 },                 // antiB-
          { 179, 5.27965 },                 // antiB0
          { 180, 5.36688 },                 // B_s0
          { 181, 5.36688 },                 // antiB_s0
          { 182, 6.2749 },                  // B_c+
          { 183, 6.2749 },                  // antiB_c-
          { 184, 5.61960 },                 // Lambda_b0
          { 185, 5.81564 },                 // Sigma_b-
          { 186, 5.81056 },                 // Sigma_b+
          { 187, 5.7919 },                  // Xi_b0
          { 188, 5.7970 },                  // Xi_b-
          { 189, 6.0461 },                  // Omega_b-
          { 190, 5.61960 },                 // antiLambda_b0
          { 191, 5.81564 },                 // antiSigma_b+
          { 192, 5.81056 },                 // antiSigma_b-
          { 193, 5.7919 },                  // antiXi_b0
          { 194, 5.7970 },                  // antiXi_b+
          { 195, 6.0461 },                  // antiOmega_b+

          { 9900, 0. },     // Cherenkov photons on particle output file
                            // (NOT there without the CERENKOV option, which )
                            // (is not used for KM3NeT CORSIKA simulations)
  };

  double Mass,EBinding,CorrMass;
  if ((CORSIKAId < 200) || (CORSIKAId == 9900)) CorrMass = Masses[CORSIKAId];      // covered above
  else {                      // nuclei
    int nZ = CORSIKAId%100;       // number of protons
    int nA = (CORSIKAId-nZ)/100;  // number of protons + neutrons
    int nN = nA - nZ;             // number of neutrons
    // without binding energy
    Mass = Masses[13] * nN + Masses[14] * nZ;

    // binding energy correction calculated from the fit to the
    // Bethe–Weizsäcker mass formula:
    // Wapstra, A. H. (1958), 'Atomic Masses of Nuclides', In Flügge, S. (ed.),
    // External Properties of Atomic Nuclei, Encyclopedia of Physics, 8/38/1,
    // Springer, pp. 1–37, Bibcode:1958HDP....38....1W,
    // doi:10.1007/978-3-642-45901-6_1, ISBN 978-3-642-45902-3
    //
    // also described in https://en.wikipedia.org/wiki/Semi-empirical_mass_formula

    EBinding = 14.1 * nZ - 0.595 * nZ*nZ * TMath::Power(nZ,-1./3.) - 13. * TMath::Power(nZ,2./3.) - 19. * (nZ-nN)*(nZ-nN) / nZ;
    
    // Here we add the pairing term:
    if ( nZ == 0 && nN > 0 ) EBinding = 0.;                                                     // multi-neutron clusters
    else if ( nA % 2 == 1 ) EBinding = EBinding;                                                // nA odd
    else if ( nZ % 2 == 0 && nN % 2 == 0 ) EBinding = EBinding + 33.5 * TMath::Power(nZ,-0.75); // nA even
    else if ( nZ % 2 == 1 && nN % 2 == 1 ) EBinding = EBinding - 33.5 * TMath::Power(nZ,-0.75); // nA even
    else {
      LOG("GSeaCORSIKAFileFlux", pFATAL) << "Unknown nucleus (A,Z): (" << nA << "," << nZ << "), exiting...";
      exit(1);
    }

    if ( EBinding > 0. ) EBinding=1.e-3*EBinding; // convert from MeV to GeV
    else EBinding=0.;

    CorrMass = Mass-EBinding;

  }
  return CorrMass;
}


double GSeaCORSIKAFileFlux::CRFlux(double PrimEnergy, int PrimCORSIKAId, string CRFluxModel){
  // implementation adapted from https://crfluxmodels.readthedocs.io/en/latest/_modules/models.html#PrimaryFlux
  // output in m^2 / {s * sr * GeV}

  double Flux = 0.0;

  if (CRFluxModel=="GST3") {

    // T. K. Gaisser, T. Stanev, and S. Tilav, arXiv:1303.3565, (2013)

    const double RigCut[3] = {120e3,4e6,1.3e9}; // Rigidity cutoffs

    std::unordered_map<int, int> NuclToID = {
      { 14, 0 },    // p
      { 402, 1 },   // He
      { 1206, 2 },  // C
      { 1608, 3 },  // O
      { 5626, 4 },  // Fe
    };

    double Params[5][3][3] = {
      { // p (14)
        {7000., 1.66, 1.},
        {150., 1.4, 1.},
        {14., 1.4, 1.}
      },{ // He (402)
        {3200., 1.58, 2.},
        {65., 1.3, 2.},
        {0., 1.4, 2.}
      },{ // C (1206)
        {100., 1.4, 6.},
        {6., 1.3, 6.},
        {0., 1.4, 6.}
      },{ // O (1608)
        {130., 1.4, 8.},
        {7., 1.3, 8.},
        {0., 1.3, 8.}
      },{ // Fe (5626)
        {60., 1.3, 26.},
        {2.3, 1.2, 26.},
        {0.025, 1.2, 26.}
      }
};

    for (int i = 0; i < 3; i++) {
      Flux += Params[NuclToID[PrimCORSIKAId]][i][0] * pow(PrimEnergy,(-Params[NuclToID[PrimCORSIKAId]][i][1] - 1.0)) * exp(-PrimEnergy / (Params[NuclToID[PrimCORSIKAId]][i][2] * RigCut[i]));
    }

  }
  else {
    LOG("GSeaCORSIKAFileFlux", pFATAL) << "Undefined primary CR flux model, exiting...";
    exit(1);
  }

  return Flux;
}
