//____________________________________________________________________________
/*!

\program gCalcAstro

\brief   Calculate astro cooridnates       

\author  Carla Distefano
         INFN - LNS

\created October 27, 2020

\cpright  Copyright (c) 2020, The KM3NeT Collaboration


*/
//____________________________________________________________________________


#include <cassert>
#include <cstdlib>
#include <cctype>
#include <string>
#include <vector>
#include <sstream>
#include <map>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <algorithm>

#include "libxml/parser.h"
#include "libxml/xmlmemory.h"
#include "libxml/xmlreader.h"

#include <TSystem.h>
#include <TRandom.h>
#include <TGraph.h>
#include <TFile.h>


// GENIE
#ifdef _GENIEVERSIONGTEQ3__
#include "Framework/Conventions/Constants.h"
#include "Framework/Conventions/Units.h"
#include "Framework/Messenger/Messenger.h"
#include "Framework/Utils/CmdLnArgParser.h"
#include "Framework/Utils/RunOpt.h"
#include "Framework/Utils/StringUtils.h"
#else
#include "Conventions/Constants.h"
#include "Conventions/Units.h"
#include "Messenger/Messenger.h"
#include "Utils/CmdLnArgParser.h"
#include "Utils/RunOpt.h"
#include "Utils/StringUtils.h"
#endif


#include "SeaNuDrivers/GAstro.h"

using namespace std;
using namespace genie;
using namespace genie::constants;

void GetCommandLineArgs (int argc, char ** argv);
void PrintSyntax        (void);

double gLatitude=0.746982;    		//  in rad correspondig to 42.798916 deg;
double gLongitude=0.107612;   		//  in rad correspondig to 6.1657 deg;
double gUTMConvergeAngle=-0.0335949;	//  in rad correspondig to -1.924846 deg;
bool   gUseUTMSystem=true;


//  Latitude=0.633407;
//  Longitude=0.278819;
//  UTMConvergeAngle=0.0100732;


string gInFileName1, gInFileName2;



double CalcDista(double theta1, double phi1, double theta2, double phi2);

double CalcUTMConvergeAngle(void); 
 

void CheckConvAngle();
 
void CheckFile1();
void CheckFile2(); 


 
const double Deg2Rad=kPi/180.;


GAstro *astro;


//___________________________________________________________________
int main(int argc, char ** argv)
{


  string filepath = gSystem->Getenv("GSEAGEN");
  filepath.append("/src/tools/gCalcAstro/");

  gInFileName1 = filepath + "antares_coordinate_systems_benchmark.csv";
  gInFileName2 = filepath + "antares_astro_objects_benchmark.csv";





  GetCommandLineArgs (argc, argv);
  
  astro = new GAstro(gLatitude, gLongitude, gUTMConvergeAngle, gUseUTMSystem);

  
  CheckConvAngle();
  
  CheckFile1();
  CheckFile2();
  
 


  return 0;
}
//___________________________________________________________________
void GetCommandLineArgs(int argc, char ** argv)
{


  LOG("gCalcAstro", pINFO) 
     << "*** Parsing command line arguments";

  LOG("gCalcAstro", pINFO) << "Parsing command line arguments";


  size_t apos;
  string opt,value;



  for (int i = 1; i < argc; i++)
  {
    string arg(argv[i]);
    

    if (arg.compare(0,1,"-")==0){
      apos=arg.find(" ");
      opt=arg.substr(0,apos);

      
      if(opt.compare("-coord")==0){      // detector coordinates
        i++;
        string coord_str = argv[i]; 
        string::size_type column = coord_str.find_first_of(":",0);  
        if(column==string::npos) {
          LOG("gCalcAstro",pFATAL) << "You need to specify the detector coordinate units: DEG or RAD ";
          gAbortingInErr = true;
          exit(1);    
        }
        string coord_unit = coord_str.substr(0,column);
        coord_str.erase(0,column+1);    
        string::size_type comma = coord_str.find_first_of(",",0);
        if(comma==string::npos){
          LOG("gCalcAstro",pFATAL) << "Error in detector coordinates input ";
          gAbortingInErr = true;
          exit(1);    
        }
        vector<string> coords = utils::str::Split(coord_str, ",");
        assert(coords.size() == 2);     
        gLatitude  = atof(coords[0].c_str());
        gLongitude = atof(coords[1].c_str());
	
        if(coord_unit.compare("DEG")==0){
	  LOG("gCalcAstro", pINFO) << "Setting detector coordinates: Latitude "<<gLatitude<<" deg; Longitude: "<<gLongitude<<" deg";
          gLatitude  *= kPi/180.;
          gLongitude *= kPi/180.;
        }
        else if(coord_unit.compare("RAD")==0){
	  LOG("gCalcAstro", pINFO) << "Setting detector coordinates: Latitude "<<gLatitude<<" rad; Longitude: "<<gLongitude<<" rad";
        }          
        else {
          LOG("gCalcAstro",pFATAL) << "Error in detector coordinate units; they must be DEG or RAD ";
          gAbortingInErr = true;
          exit(1);    
        }          
      }
      
      
      if(opt.compare("-alpha")==0){      // convergence angle
        i++;
        string alpha_str = argv[i]; 
      
        string::size_type column = alpha_str.find_first_of(":",0);  
        if(column==string::npos) {
          LOG("gCalcAstro",pFATAL) << "You need to specify the convergence angle unit: DEG or RAD ";
          gAbortingInErr = true;
          exit(1);    
        }
        string alpha_unit = alpha_str.substr(0,column);
        alpha_str.erase(0,column+1);    

        gUTMConvergeAngle = atof(alpha_str.c_str());

        if(alpha_unit.compare("DEG")==0){
	  LOG("gCalcAstro", pINFO) << "Setting UTM convergence angle: "<<gUTMConvergeAngle<<" deg";
          gUTMConvergeAngle  *= kPi/180.;
        }
        else if(alpha_unit.compare("RAD")==0){
	  LOG("gCalcAstro", pINFO) << "Setting UTM convergence angle: "<<gUTMConvergeAngle<<" rad";
        }          
        else {
          LOG("gCalcAstro",pFATAL) << "Error in UTM convergence angle unit; it must be DEG or RAD ";
          gAbortingInErr = true;
          exit(1);    
        }          

      }
      
      
      if(opt.compare("-file1")==0){      // input file 1: from track to Galactic coord
        i++;
        gInFileName1 = argv[i]; 
	LOG("gCalcAstro", pINFO) << "Setting input file 1: "<<gInFileName1;	
      }	      

      if(opt.compare("-file2")==0){      // input file 2: from Equatorial J2000 to track  
        i++;
        gInFileName2 = argv[i]; 
	LOG("gCalcAstro", pINFO) << "Setting input file 2: "<<gInFileName2;

      }	      
      

      if(opt.compare("-utm")==0){      // check if utm or gSeaGen reference system
        i++;
        gUseUTMSystem = atoi(argv[i]); 
	if(gUseUTMSystem) LOG("gCalcAstro", pINFO) << "Setting UTM reference system";
	else LOG("gCalcAstro", pINFO) << "Setting gSeaGen reference system";
      }	      

      if(opt.compare("-h")==0){         // help
        PrintSyntax();
        exit(0);
      }      
    }
  }


  //
  // print-out summary
  //


  LOG("gCalcAstro", pINFO)
       << "\n"
       << "\n ****** Input summary ****** \n"
       << "\n @@ Detector coordinates"
       << "\n\t Latitude:  " << gLatitude  <<" rad \t\t ("<<gLatitude*180./kPi<<" deg)"
       << "\n\t Longitude: " << gLongitude <<" rad \t\t ("<<gLongitude*180./kPi<<" deg)"

       << "\n @@ Coordinate system";   
       if(gUseUTMSystem){
         cout<< "\t UTM system\n";           
         cout<< "\t Convergence angle:  " << gUTMConvergeAngle  <<" rad \t ("<<gUTMConvergeAngle*180./kPi<<" deg)";
       }
       else {
         cout<< "\t gSeaGen system";
       }
       cout<< "\n\n";

}

//_________________________________________________________________________________
void PrintSyntax(void)
{

  LOG("gCalcAstro", pFATAL)
     << "\n\n"
     << "gCalcAstro                  	\n"
/*
     << "    [-e emin,emax]	     	\n"
     << "    [-n nutype] 	   	\n"
     << "    [-b nbins] 		\n"
     << "    [-f xsec.xml]              \n"
     << "    [-i interaction] 		\n"
#ifdef _GENIEVERSIONGTEQ3__     
     << "    [-m genie_tune] 		\n"     
#endif         
     << "    [-t target.xml]            \n"     
     << "    [-o output.root]           \n"
     << "    [-h]	 		\n"
     << "\n"
     << "\n **Code Options**"      
     << "\n"   
     << "\n [] Denotes an optional argument."
     << "\n"    
     << "\n -e        Specifies the neutrino energy in GeV [default: 1,5000]. " 
     << "\n           Must be a comma-separated pair of numbers, eg. `-e 100,500' " 
     << "\n -n        Specifies the neutrino type in the PDG code (12: nu_e, -12: anti_nu_e, 14: nu_mu,"
     << "\n           -14: anti_nu_mu, 16: nu_tau, -16: anti_nu_tau) [default: 14]. "
     << "\n"
     << "\n -b        Specifies the numebr of energy bins for xsec calculation [default: 100]. "
     << "\n"
     << "\n -f        Specifies the input XML cross section spline file. [default: '$GSEAGEN/dat/gxspl-seawater-muon.xml']"   
     << "\n"
     << "\n -i        Interaction type. Valid settings are the XML block names appearing in"   
     << "\n           config/EventGeneratorListAssembler.xml [default: 'Default']."
     << "\n"
     << "\n           eg. '-i  Default' (the default considers all interaction types)"
     << "\n           eg. '-i  CC' "
     << "\n           eg. '-i  NC' "   
     << "\n"
#ifdef _GENIEVERSIONGTEQ3__
     << "\n -m        Specifies the genie tune [default: G18_02a_00_000]. "
     << "\n"
     << "\n           eg. '-t G18_02a_00_000'" 
     << "\n"    
#endif     
     << "\n -t        Specifies file name containing the target mixture [default: water]. "
     << "\n"
     << "\n           eg. '-t $GSEAGEN/src/tools/gCalcCross/water.xml'"     
     << "\n"   
     << "\n -o        Specifies the output file name containing the cross sections [default: xsec.root]. "
     << "\n"
*/
     << "\n -h        Prints this help"
     << "\n";

     
}

//_________________________________________________________________________________
double CalcDista(double theta1, double phi1, double theta2, double phi2){
  double distance;

  double vx1,vy1,vz1;
  double vx2,vy2,vz2;
  
  vx1=-sin(theta1)*cos(phi1);
  vy1=-sin(theta1)*sin(phi1);
  vz1=-cos(theta1);
  
  vx2=-sin(theta2)*cos(phi2);
  vy2=-sin(theta2)*sin(phi2);
  vz2=-cos(theta2);
  
  
  
  
  double unit1=sqrt(vx1*vx1+vy1*vy1+vz1*vz1);
  double unit2=sqrt(vx2*vx2+vy2*vy2+vz2*vz2);

  
  distance=vx1*vx2+vy1*vy2+vz1*vz2;
  distance=distance/(unit1*unit2);
  distance=acos(distance)/Deg2Rad;


  return distance;
}

//_________________________________________________________________________________
void CheckConvAngle(){

// Check calculation of convergence angle
  
  cout<<endl<<endl;
  cout << "Check calculation of convergence angle"<<endl;

  double UTMConvergeAngle = CalcUTMConvergeAngle();
  

  
  printf("\nConvergence angle: \t %f rad \t (%f deg)\n",UTMConvergeAngle, UTMConvergeAngle/Deg2Rad);
  printf("Error: \t\t\t %f rad (%f deg)\n\n",(UTMConvergeAngle-UTMConvergeAngle),(UTMConvergeAngle-UTMConvergeAngle)/Deg2Rad);

}
//_________________________________________________________________________________
void  CheckFile1(){

  string date,time;
  string DataString;
  
  double Theta_In,Phi_In,Zenith_In,Azimuth_In,DEC_In,RA_In,DEC_J2000_In,RA_J2000_In,GalLat_In,GalLong_In;

  double LST;
  double MJD;
 
  double Azimuth,Altitude,Zenith;
  double GalLat,GalLong;
  double RA,Dec;
  double vx,vy,vz;  
  
  string DecStr,RAStr;
  vector<string> Input,Output;

  printf("\nFrom track to Galactic coord\n\n");

    
  ifstream file(gInFileName1.c_str());
  
  if(file.is_open())
    printf("Input file: %s\n\n",gInFileName1.c_str());
  else{
    printf("Error in opening input file: %s\n\n",gInFileName1.c_str());  
    return;
  }
  

  char buffer [1024];  
  
  string line;
  while (file.good()) {
    getline(file, line);
    if(line.find("#")== string::npos)break; 
  }

  
  while (file>>line){
    std::replace( line.begin(), line.end(), ',', ' '); 
 
    istringstream iline (line);
    iline>>date>>time>>Theta_In>>Phi_In>>Zenith_In>>Azimuth_In>>DEC_In>>RA_In>>DEC_J2000_In>>RA_J2000_In>>GalLat_In>>GalLong_In;
  
    DataString=date.substr(8,2)+date.substr(4,3)+"-"+date.substr(0,4)+";"+time;

    MJD=astro->CalcMJD(DataString); 
    LST=astro->LST(MJD);
    
    vx=-sin(Theta_In*Deg2Rad)*cos(Phi_In*Deg2Rad);
    vy=-sin(Theta_In*Deg2Rad)*sin(Phi_In*Deg2Rad);
    vz=-cos(Theta_In*Deg2Rad);

    astro->Direction2Hor(vx,vy,vz,&Azimuth,&Altitude);
     
    double RA0,Dec0; 
    astro->Hor2Eq(Altitude,Azimuth, LST, &RA0, &Dec0);
    
    astro->CalcPrecession(MJD, 51544.5, RA0, Dec0, &RA, &Dec);
    
    astro->Eq2Gal("J2000", RA, Dec, &GalLat, &GalLong);

    Azimuth/=Deg2Rad;
    Altitude/=Deg2Rad;

    Zenith=90.-Altitude;

    Dec/=Deg2Rad;
    RA/=Deg2Rad;
    GalLat/=Deg2Rad;
    GalLong/=Deg2Rad;
    
    sprintf(buffer,"%s %s %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f ",date.c_str(),time.c_str(),Theta_In,Phi_In,Zenith_In,Azimuth_In,DEC_J2000_In,RA_J2000_In,GalLat_In,GalLong_In);
    Input.push_back(buffer);
	

    sprintf(buffer,"%s %s %f \t %f \t %f \t %f \t %f \t %f \t %f \t %f \t | \t %f \t %f \t %f \t %f",date.c_str(),time.c_str(),Theta_In,Phi_In,Zenith,Azimuth,Dec,RA,GalLat,GalLong,Dec-DEC_J2000_In,RA-RA_J2000_In,GalLat-GalLat_In,GalLong-GalLong_In);
    Output.push_back(buffer);


  
  }
  
  printf("Date \t\t    Theta\t Phi \t\t Zenith \t Azimuth  \t Dec (J2000) \t RA (J2000) \t GalLat \t GalLong \t \n");

  for(unsigned i=0; i<Input.size(); i++)cout<<Input[i]<<endl;
  cout<<endl;

  printf("Calculated coordinates \n\n");
  printf("Date \t\t    Theta\t Phi \t\t Zenith \t Azimuth  \t Dec \t\t RA \t\t GalLat \t GalLong \t | \t ErrDec \t ErrRA \t\t ErrGalLat \t ErrGalLong \n");
  for(unsigned i=0; i<Output.size(); i++)cout<<Output[i]<<endl;
  
  
  file.close();
  
  return;
}  


//_________________________________________________________________________________

void  CheckFile2(){


  string date,time;
  string DataString;
  
  double Theta_In,Phi_In,Zenith_In,Azimuth_In,DEC_J2000_In,RA_J2000_In;

  double LST;
  double MJD;
 
 
  double Theta,Phi;
  double Azimuth,Altitude,Zenith;
  double RA,Dec;
  double vx,vy,vz;  
  
  string DecStr,RAStr;
  int var1,var2,var3;

  vector<string> Input,Output;

  printf("\n\n\nFrom Equatorial J2000 to track \n\n");

    
  ifstream file(gInFileName2.c_str());
  
  if(file.is_open())
    printf("Input file: %s\n\n",gInFileName2.c_str());
  else{
    printf("Error in opening input file: %s\n\n",gInFileName2.c_str());  
    return;
  }

  string line;
  while (file.good()) {
    getline(file, line);
    if(line.find("#")== string::npos)break; 
  }

  string name;
  
  char buffer [1024];
  
  while (file>>line){
    std::replace( line.begin(), line.end(), ',', ' '); 
 
    istringstream iline (line);
    
    iline>>date>>time>>name>>RAStr>>DecStr>>Zenith_In>>Azimuth_In>>Theta_In>>Phi_In;
  
    DataString=date.substr(8,2)+date.substr(4,3)+"-"+date.substr(0,4)+";"+time;

    MJD=astro->CalcMJD(DataString); 
    LST=astro->LST(MJD);

    
    std::replace( RAStr.begin(), RAStr.end(), ':', ' '); 
    std::replace( DecStr.begin(), DecStr.end(), ':', ' '); 

    istringstream iRAStr (RAStr);
    istringstream iDecStr (DecStr);
    
    iRAStr>>var1>>var2>>var3;    
    RA_J2000_In= ((double)var1+(double)(var2)/60.+(double)var3/3600.)*15.;
    
    iDecStr>>var1>>var2>>var3;    
    DEC_J2000_In= (var1/fabs(var1))*(fabs((double)var1)+(double)(var2)/60.+(double)var3/3600.);
    
    astro->CalcPrecession(51544.5, MJD, RA_J2000_In*Deg2Rad, DEC_J2000_In*Deg2Rad, &RA, &Dec);
    astro->CalcNutation(MJD, RA, Dec, &RA, &Dec);
    astro->Eq2Hor(LST, RA, Dec, &Azimuth,&Altitude);
    astro->Hor2Direction(Azimuth,Altitude, &vx,&vy,&vz);

    Azimuth/=Deg2Rad;
    Altitude/=Deg2Rad;
    Zenith=90.-Altitude;

    Theta=acos(-vz)/Deg2Rad;
    Phi=atan2(-vy,-vx)/Deg2Rad;
    Phi=fmod(Phi+360.,360.);

    astro->CalcSrcDirMJD(MJD, RA_J2000_In*Deg2Rad, DEC_J2000_In*Deg2Rad, &vx, &vy, &vz);
    
//    double Theta1=acos(-vz)/Deg2Rad;
    double Phi1=atan2(-vy,-vx)/Deg2Rad;
    Phi1=fmod(Phi1+360.,360.);

    sprintf(buffer,"%s %s %s \t %s \t %s \t %f \t %f \t %f \t %f",date.c_str(),time.c_str(),name.c_str(),RAStr.c_str(),DecStr.c_str(),Zenith_In,Azimuth_In,Theta_In,Phi_In);
    Input.push_back(buffer);
    
    sprintf(buffer,"%s %s %s \t %s \t %s \t %f \t %f \t %f \t %f \t | \t %f \t %f ",date.c_str(),time.c_str(),name.c_str(),RAStr.c_str(),DecStr.c_str(),Zenith,Azimuth,Theta,Phi,Theta-Theta_In,Phi-Phi_In);
    Output.push_back(buffer);
  }

  printf("Date \t\t    Object \t RA \t\t Dec \t\t Zenith \t Azimuth \t Theta \t\t Phi \t\t \n");
  for(unsigned i=0; i<Input.size(); i++)cout<<Input[i]<<endl;
  cout<<endl;

  printf("Calculated coordinates \n\n");
  printf("Date \t\t    Object \t RA \t\t Dec \t\t Zenith \t Azimuth \t Theta \t\t Phi \t\t | \t ErrTheta \t ErrPhi \n");
  for(unsigned i=0; i<Output.size(); i++)cout<<Output[i]<<endl;

  file.close();

  
  return;
}  

//___________________________________________________________________________
double CalcUTMConvergeAngle(void){

  // Calculate the UTM convergence angle

  double zone=1+int( 60.* (  gLongitude*180./kPi +180.)/360.);
  double lambda0=(-180.+(zone-1.)*6.+3)*kPi/180.; // in rad

  double lambda  = gLongitude;
  double phi     = gLatitude;
  double omega   = lambda-lambda0;

  // parameters of the Earth ellipsoid
  double a  = 6378137.;      // semi-major axis in meters (WGS84)
  double e2 = 0.0066943800;  // eccentricity (WGS84)

  double rho = a*(1.-e2)/pow(1.-e2*pow(sin(phi),2),3./2.);
  double nu  = a/sqrt(1-e2*pow(sin(phi),2.));
  double psi = nu/rho;
  double t = tan(phi);

  // power series for convergence angle
  double UTMConvergeAngle = sin(phi) * omega
    - sin(phi) * pow(omega,3.)/3.  * pow(cos(phi),2.) * (2.*pow(psi,2.)-psi)
    - sin(phi) * pow(omega,5.)/15. * pow(cos(phi),4.) * (pow(psi,4.)*(11.-24.*pow(t,2.))-
							 pow(psi,3.)*(11.-36.*pow(t,2.))+
							 2.*pow(psi,2.)*(1.-7.*pow(t,2.))+
							 psi*pow(t,2.)
							 )
    - sin(phi) * pow(omega,7.)/315. * pow(cos(phi),6.) * (17.-26.*pow(t,2.)+2.*pow(t,4.));

  return UTMConvergeAngle;
}

